#include <stdio.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <signal.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#include "tree.h"
#include "dialitd.h"
#include "commands.h"

config_t * read_config(int argc,char** argv,char** envp)
{
    int loop;
    config_t * config;

    config = malloc(sizeof(config_t));
    if (config==NULL)
    {
	fprintf(stderr,"Unable to allocate store for config\n");
	return NULL;
    }

    /* set our quit status */
    config->quit=0;

    config->waittime=1000;

    for (loop=0;loop < OPEN_MAX;loop++)
    {
	config->client[loop].fd = -1;
	config->client_data[loop].line_buffer=malloc(LINE_BUFFER_LEN);
	config->client_data[loop].buffer_end=0;
    }

    config->maxclient=0;

    config->treeroot = make_tree_list("/etc/ppp/peers");

    if(config->treeroot == NULL)
    {
	free(config);
	return NULL;
    }

    config->orig_cmd_line=strdup(argv[0]);

    return config;
}

int create_listener(config_t * config)
{
    struct sockaddr_in serv_addr;
    int sockfd;

    sockfd=socket(AF_INET,SOCK_STREAM,0);
    if(sockfd<0)
    {
	fprintf(stderr,"Unable to open socket\n");
	return 1;
    }

    bzero((char *)&serv_addr,sizeof(serv_addr));
    serv_addr.sin_family=AF_INET;
    serv_addr.sin_addr.s_addr=htonl(INADDR_ANY);
    serv_addr.sin_port=htons(SERV_TCP_PORT);

    if (bind(sockfd,(struct sockaddr *) &serv_addr,sizeof(serv_addr))<0)
    {
	fprintf(stderr,"Unable to bind to local address\n");
	close(sockfd);
	return 1;
    }

    if(listen(sockfd,5) < 0)
    {
	fprintf(stderr,"Unable to listen on socket\n");
	close(sockfd);
	return 1;
    }

    config->client[0].fd=sockfd;
    config->client[0].events=POLLIN;

    return 0;
}

int demonise()
{
    int childpid;

    if ((childpid = fork()) <0)
    {
	fprintf(stderr,"Error cannot fork first child");
	return 1;
    }
    else if(childpid >0)
	exit(0); /* ditch controling parent */


    if (setpgrp() == -1)
	fprintf(stderr,"error cannot change process group");

    signal(SIGHUP,SIG_IGN);


    if ((childpid = fork()) <0)
    {
	printf("Error cannot fork second child");
	return 1;
    }
    else if(childpid >0)
	exit(0); /* ditch controling parent */

    freopen("/dev/null","r",stdin);
    freopen("/dev/null","w",stdout);
    freopen("/dev/null","w",stderr);

    chdir("/");

    umask(0);
    return 0;
}

int send_buffer(int fd,char * data,int data_len,int retries)
{
    int sent_len;
    int sent_reply=0;

    sent_reply=write(fd,data,data_len);
    if(sent_reply<0)
    {
	perror("dialitd: send_reply");
	return -1;
    }
    else
	sent_len=sent_reply;

    while(sent_len<data_len)
    {
	sent_reply=write(fd,&data[sent_len],data_len-sent_len);
	if(sent_reply<0)
	{
	    perror("dialitd: send_reply");
	    return sent_len;
	}
	else
	    sent_len+=sent_reply;
    }
    return sent_len;

}

int send_reply(int fd,int code,char * message,...)
{
    int reply_len;
    char txt_buffer[4096];
    va_list argp;

    reply_len=snprintf(txt_buffer,4096,"%d%c",abs(code),(code<0?' ':'+'));
    if(send_buffer(fd,txt_buffer,reply_len,20)<0)
	return 1;

    va_start(argp, message);
    reply_len=vsnprintf(txt_buffer,4096,message,argp);
    va_end(argp);

    if(reply_len<1)
	return 1;

    if(send_buffer(fd,txt_buffer,reply_len,20)<0)
	return 1;

    return 0;
}


int new_connection(config_t * config)
{
    struct sockaddr_in cli_addr;
    int clilen;
    int newfd;
    int loop;

    clilen=sizeof(cli_addr);
    newfd=accept(config->client[0].fd,(struct sockaddr*) &cli_addr,&clilen);
    if(newfd<0)
    {
	/* error in accept */
	perror("dialitd");
	return 1;
    }

    for(loop=1;loop < OPEN_MAX;loop++)
    {
	if(config->client[loop].fd <0)
	{
	    config->client[loop].fd = newfd;
	    config->client[loop].events = POLLIN;
	    if(loop > config->maxclient)
		config->maxclient = loop;

	    send_reply(newfd,200,"Dialitd Version "VERSION"\n");
	    send_reply(newfd,-200,"Connection Open - Waiting commands\n");

	    break;
	}
    }

    /* check we havnt reached our connection limit */
    if(loop==OPEN_MAX)
    {
	close(newfd);
	return 1;
    }
    return 0;

}

int my_strnlen(char * inp_str,int max_len)
{
    int loop;
    for(loop=0;loop<=max_len;loop++)
	if(inp_str[loop]==0)
	    break;

    return loop;
}

char * my_strndup(char * inp_str,int max_len)
{
    char * new_str;
    int inp_str_len;

    inp_str_len=my_strnlen(inp_str,max_len);

    if(inp_str_len<max_len)
	max_len=inp_str_len;

    new_str=malloc(max_len+1);

    strncpy(new_str,inp_str,max_len);

    new_str[max_len]=0;

    return new_str;

}


char * read_line(client_data_st * client_data)
{
    char * out_line;
    int loop;
    int buffer_remain=0;
    int line_end;

    /* empty buffer - no command */
    if(client_data->buffer_end == 0)
	return NULL;

    /* find newline to terminate line */
    for(loop=0;loop <= client_data->buffer_end;loop++)
	if((client_data->line_buffer[loop] == '\r') ||
	   (client_data->line_buffer[loop] == '\n'))
	    break;

    line_end=loop;

    /* didnt find newline in available data */
    if(loop > client_data->buffer_end)
    {
	if(client_data->buffer_end == LINE_BUFFER_LEN)
	{
	    /* full buffer but no newline just return the whole buffer */
	    out_line=client_data->line_buffer;
	    client_data->line_buffer=malloc(LINE_BUFFER_LEN);
	    client_data->buffer_end=0;
	}
	else
	    return NULL;
    }
    else
    {
	/* skip any junk newlines */
	for(;loop <= client_data->buffer_end;loop++)
	    if((client_data->line_buffer[loop] != '\r') &&
	       (client_data->line_buffer[loop] != '\n'))
		break;

	/* copy the line to return (omit the newline) */
	out_line=my_strndup(client_data->line_buffer,line_end);

	buffer_remain=(client_data->buffer_end - loop);

	if (buffer_remain>0)
	{

	    /* move any extra data down the buffer to the start */
	    memmove(client_data->line_buffer,
		    &client_data->line_buffer[loop],
		    buffer_remain);

	    /* reset the line length acordingly */
	    client_data->buffer_end = buffer_remain;
	}
	else
	    client_data->buffer_end = 0;

    }
    return out_line;
}

/* reads data into buffer and returns the total amount available in the buffer
 *  except for errors or eof!
 */
int read_buffered(struct pollfd *client, client_data_st * client_data)
{
    int amntread=0;
    int buffer_avail=0;

    buffer_avail=LINE_BUFFER_LEN - client_data->buffer_end;

    if(buffer_avail > 0)
    {
	amntread=read(client->fd,
		      &client_data->line_buffer[client_data->buffer_end]
		      ,buffer_avail);

	if(amntread <= 0)
	    return amntread;
    }

    client_data->buffer_end+=amntread;

    return client_data->buffer_end;
}

int close_client(config_t * config,int client_number)
{
    if(close(config->client[client_number].fd) < 0)
    {
	send_reply(config->client[client_number].fd,-500,"Error In close\n");
	return 1;
    }

    config->client[client_number].fd=-1;
    config->client_data[client_number].buffer_end=0;
    return 0;
}




int service_client(config_t * config,int client_number)
{
    int amntread;
    char *cmnd_ln;

    amntread=read_buffered(&config->client[client_number],
			   &config->client_data[client_number]);
    if(amntread <0)
    {
	if (errno == ECONNRESET)
	    close_client(config,client_number);
	else
	{
	    /*  error */
	}
    }
    else if(amntread==0)
	close_client(config,client_number);
    else
    {
     	while((cmnd_ln = read_line(&config->client_data[client_number]))!=NULL)
	    command(config,cmnd_ln,client_number);

    }

    return 0;

}

int close_sockets(config_t * config)
{
    int loop;

    for (loop=0;loop<=config->maxclient;loop++)
    {
	/* checks its a valid fd */
	if (config->client[loop].fd < 0)
	    continue;

	close_client(config,loop);
    }

    return 0;
}


int service_socket(config_t * config)
{
    int nready;/* number of ready sockets */
    int loop;

    nready=poll(config->client,config->maxclient+1,config->waittime);

    if (nready==0)
	return 0; /* timed out */
    else if(nready<0)
    {
	perror("dialitd");
	return 1;
    }

    /* check for new connection */
    if (config->client[0].revents & POLLIN)
    {
	/* accept new connection */
	new_connection(config);

	if (--nready <=0)
	    return 0;/* no more descriptors ready */
    }

    /* check for action on client connections */
    for (loop=1;loop<=config->maxclient;loop++)
    {
	/* checks its a valid fd */
	if (config->client[loop].fd < 0)
	    continue;

	/* find if socket has event pending */
	if(config->client[loop].revents & (POLLIN | POLLERR))
	{
	    service_client(config,loop);

	    if (--nready <=0)
		return 0;/* no more descriptors ready */
	}

    }
    return 0;
}

int check_link_status(config_t * config)
{
    int result=0;
    int sockfd;
    struct ifreq ifr;

    /* make a socket for the ioctl */
    if((sockfd=socket(AF_INET,SOCK_DGRAM,0)) < 0)
    {
	return 0;
    }

    /* call the ioctl and process return condition */
    strncpy(ifr.ifr_name, "ppp0", IFNAMSIZ);

    if((ioctl(sockfd,SIOCGIFFLAGS,&ifr) >= 0) &&
       ((ifr.ifr_flags & IFF_UP) != 0))
    {
	result=1;
    }
    else
    {
	/* might be dialing */
	if(pppd_kill("ppp0",0)==0)
	    result=2;

    }

    close(sockfd);

    return result;

}

int update_status(config_t * config)
{
    if(config->quit)
	return 1;

    config->link_active=check_link_status(config);

    return 0;
}

int main(int argc,char** argv,char** envp)
{
    config_t * config;

    /* read config and exit if problem */
    config=read_config(argc,argv,envp);
    if(config==NULL)
	return 1;

    /* create listening socket */
    if(create_listener(config))
	return 2;

    /* make us a background process */
    if (demonise())
	return 3;

    /* update our status for ever */
    while(!update_status(config))
    {
	/* service background listening socket - blocks for a time*/
	service_socket(config);
    }

    close_sockets(config);

    free_tree_list(config->treeroot);

    /* were done */
    return 0;

}
